import * as THREE from '../build/three.module.js';

import Stats from './jsm/libs/stats.module.js'
import { GUI } from './jsm/libs/dat.gui.module.js'

import { OrbitControls } from './jsm/controls/OrbitControls.js'
import { TransformControls } from './jsm/controls/TransformControls.js'
 
class App {
    constructor() {
        this._initialize();
    }

    _initialize() {
        this.domWebGL = document.createElement('div');
        document.body.appendChild(this.domWebGL);

        let scene = new THREE.Scene();
        scene.background = new THREE.Color(0xf0f0f0);

        let renderer = new THREE.WebGLRenderer( { antialias: true });
        renderer.shadowMap.enabled = true;
        
        this.domWebGL.appendChild(renderer.domElement);

        this._stats = new Stats();
        this.domWebGL.appendChild(this._stats.dom);

        this._setupGUI();    

        window.onresize = this.resize.bind(this);

        this.renderer = renderer;
        this.scene = scene;

        this.ARC_SEGMENTS = 100;

        this._splineHelperObjects = [];
        this._positions = [];
        this._splines = {};

        this._markerGeometry = new THREE.BoxGeometry(20, 20, 20);
        this._markerMaterial = new THREE.MeshLambertMaterial({color:0x777777});

        this._onUpPosition = new THREE.Vector2();
        this._onDownPosition = new THREE.Vector2();
        this._raycaster = new THREE.Raycaster();

        this._setupModel();
        this._setupLights()
        this._setupCamera();
        this._setupControl();

        this.resize();
    }

    _setupGUI() {
        const gui = new GUI();

        const params = {
            uniform: true,
            tension: 0.5,
            centripetal: true,
            chordal: true,
            addPoint: this._addPoint.bind(this),
            removePoint: this._removePoint.bind(this)
        };

        gui.add(params, 'uniform');
        gui.add(params, 'tension', 0, 1).step(0.01).onChange((value) => {
            this._splines.uniform.tension = value;
            this._updateSplineOutline();
        });

        gui.add(params, 'centripetal');
        gui.add(params, 'chordal');
        gui.add(params, 'addPoint');
        gui.add(params, 'removePoint');
        gui.open();

        this._gui = gui;
        this._params = params;
    }

    _setupControl() {
        const orbitControls = new OrbitControls(this.camera, this.renderer.domElement);
        orbitControls.damping = 0.2;
        
        const transformControls = new TransformControls(this.camera, this.renderer.domElement);
        transformControls.addEventListener('dragging-changed', (e) => {
            orbitControls.enabled = !e.value;
        });

        this.scene.add(transformControls);
        transformControls.addEventListener('objectChange', () => {
            this._updateSplineOutline();
        })

        document.addEventListener('pointerdown', this.onPointerDown.bind(this));
        document.addEventListener('pointerup', this.onPointerUp.bind(this));
        document.addEventListener('pointermove', this.onPointerMove.bind(this));

        this._transformControls = transformControls;
    }

    _setupModel() {
        const planeGeometry = new THREE.PlaneGeometry(2000, 2000);
        planeGeometry.rotateX(-Math.PI / 2);

        const planeMaterial = new THREE.ShadowMaterial({ opacity: 0.2 });
        const plane = new THREE.Mesh(planeGeometry, planeMaterial);
        plane.position.y = -200;
        plane.receiveShadow = true;
        this.scene.add(plane);

        const helper = new THREE.GridHelper(2000, 100);
        helper.position.y = -199;
        helper.material.opacity = 0.25;
        helper.material.transparent = true;
        this.scene.add(helper);

        const positions = [
            new THREE.Vector3(289.76843686945404, 452.51481137238443, 56.10018915737797),
            new THREE.Vector3(-53.56300074753207, 171.49711742836848, -14.495472686253045),
            new THREE.Vector3(-91.40118730204415, 176.4306956436485, -6.958271935582161),
            new THREE.Vector3(-383.785318791128, 491.1365363371675, 47.869296953772746)
        ];

        for(let i=0; i<positions.length; i++) {
            this._addSplineHelperObject(positions[i]);
            this._positions.push(this._splineHelperObjects[i].position);
        }

        const geometry = new THREE.BufferGeometry();
        geometry.setAttribute('position', new THREE.BufferAttribute(new Float32Array(this.ARC_SEGMENTS*3), 3));

        let curve = new THREE.CatmullRomCurve3(this._positions);
        curve.curveType = 'catmullrom';
        curve.mesh = new THREE.Line(geometry.clone(), new THREE.LineBasicMaterial({
            color: 0xff0000, opacity: 0.35
        }));

        curve.mesh.castShadow = true;
        this._splines.uniform = curve;

        curve = new THREE.CatmullRomCurve3(this._positions);
		curve.curveType = 'centripetal';
		curve.mesh = new THREE.Line(geometry.clone(), new THREE.LineBasicMaterial({
		    color: 0x00ff00, opacity: 0.35
		}));
		
        curve.mesh.castShadow = true;
		this._splines.centripetal = curve;

        curve = new THREE.CatmullRomCurve3(this._positions);
        curve.curveType = 'chordal';
        curve.mesh = new THREE.Line( geometry.clone(), new THREE.LineBasicMaterial({
            color: 0x0000ff, opacity: 0.35
        }));

        curve.mesh.castShadow = true;
        this._splines.chordal = curve;

        for(const k in this._splines) {
            const spline = this._splines[k];
            this.scene.add(spline.mesh);
        }

        this._updateSplineOutline();
    }

    _addSplineHelperObject(position) {
        const object = new THREE.Mesh(this._markerGeometry, this._markerMaterial);

        if(position) {
            object.position.copy(position);
        } else {
            object.position.x = Math.random() * 1000 - 500;
            object.position.y = Math.random() * 600;
            object.position.z = Math.random() * 800 - 400;
        }

        object.castShadow = true;
        object.receiveShadow = true;
        this.scene.add(object);
        this._splineHelperObjects.push(object);
        
        return object;
    }

    _addPoint() {
        this._positions.push(this._addSplineHelperObject().position);
        this._updateSplineOutline();
    }

    _removePoint() {
        if(this._splineHelperObjects.length <= 4) {
            return;
        }

        const point = this._splineHelperObjects.pop();
        this._positions.pop();

        if(this._transformControls.object === point) this._transformControls.detach();
        this.scene.remove(point);

        this._updateSplineOutline();
    }

    _updateSplineOutline() {
        const point = new THREE.Vector3();
 
        for(const k in this._splines) {
            const spline = this._splines[k];
            const splineMesh = spline.mesh;
            const position = splineMesh.geometry.attributes.position;

            for(let i=0; i<this.ARC_SEGMENTS; i++) {
                const t = i / (this.ARC_SEGMENTS - 1);
                spline.getPoint(t, point);
                position.setXYZ(i, point.x, point.y, point.z);
            }

            position.needsUpdate = true;
        }
    }

    update(time) {
        this._stats.update();
    }

    render(time) {
        this.update(time);
        
        this._splines.uniform.mesh.visible = this._params.uniform;
        this._splines.centripetal.mesh.visible = this._params.centripetal;
        this._splines.chordal.mesh.visible = this._params.chordal;
        
        this.renderer.render(this.scene, this.camera);
        
        requestAnimationFrame(this.render.bind(this));
    }

    onPointerDown(event) {
        this._onDownPosition.x = event.clientX;
        this._onDownPosition.y = event.clientY;
    }

    onPointerUp(event) {
        this._onUpPosition.x = event.clientX;
        this._onUpPosition.y = event.clientY;

        if ( this._onDownPosition.distanceTo(this._onUpPosition) === 0 ) this._transformControls.detach();
    }

    onPointerMove(event) {
        if(this._transformControls.dragging) return;

        const pointer = new THREE.Vector2();
        pointer.x = (event.clientX / window.innerWidth) * 2 - 1;
        pointer.y = - (event.clientY / window.innerHeight) * 2 + 1;

        this._raycaster.setFromCamera(pointer, this.camera);

        const intersects = this._raycaster.intersectObjects(this._splineHelperObjects);

        if (intersects.length > 0) {
            const object = intersects[0].object;
            if (object !== this._transformControls.object) {
                this._transformControls.attach(object);
            }
        }
    }

    _setupLights() {
        this.scene.add(new THREE.AmbientLight(0xf0f0f0));

        const light = new THREE.SpotLight(0xffffff, 1.5);
        light.position.set(0, 1500, 200);
        light.angle = Math.PI*0.2;
        light.castShadow = true;
        light.shadow.camera.near = 200;
        light.shadow.camera.far = 2000;
        light.shadow.bias = -0.000222;
        light.shadow.mapSize.width = 1024;
        light.shadow.mapSize.height = 1024;
        
        this.scene.add(light);
    }

    _setupCamera() {
        let camera = new THREE.PerspectiveCamera(70, window.innerWidth / window.innerHeight, 1, 10000);

        camera.position.set(0, 250, 1000);

        this.scene.add(camera);
        this.camera = camera;
    }

    resize() {
        let camera = this.camera;
        let renderer = this.renderer;

        renderer.setPixelRatio(window.devicePixelRatio);
        renderer.setSize(window.innerWidth, window.innerHeight);

        camera.aspect = window.innerWidth / window.innerHeight;
        camera.updateProjectionMatrix();
    }
}

window.onload = function () {
    (new App()).render();
}